%{
Bubble Analysis

Usage:
1. set directory which contains the folder of images to analyze
2. set the file which is the folder to analyze

%}

%% Setup

clc
clear
close all

tic

main_directory = 'E:\FTO Resistor Characterization Femtosecond Curved Channels (5-8-22)\Closed Reservoir';
calibration_directory = 'C:\Users\bshay\Documents\Grad School\M-Spice PhD Project\Fabrication of Low-Cost Thermal Bubble-Driven Micro-Pumps\Calibration Images';
noise_filter_type = 'wiener'; %wiener or deep (wiener way faster)

% go through repeats

all_folders = dir(main_directory); %get all file contents
all_folders = all_folders([all_folders.isdir]); %only take those that are directories
real_folders = false(1,length(all_folders));
for ii = 1:length(all_folders)
    if length(all_folders(ii).name) < 3 || ~isempty(strfind(all_folders(ii).name,'processed')) %catches the . and .. directory links; ignored processed directories
        real_folders(ii) = false;
    else
        real_folders(ii) = true;
    end
end
real_folders_repeats = all_folders(real_folders); %filtered folder names so all voltage levels

for iter = 1:length(real_folders_repeats)

directory = fullfile(main_directory,real_folders_repeats(iter).name); %index into the repeat directory

all_folders = dir(directory); %get all file contents
all_folders = all_folders([all_folders.isdir]); %only take those that are directories
real_folders = false(1,length(all_folders));
for ii = 1:length(all_folders)
    if length(all_folders(ii).name) < 3 || ~isempty(strfind(all_folders(ii).name,'processed')) %catches the . and .. directory links; ignored processed directories
        real_folders(ii) = false;
    else
        real_folders(ii) = true;
    end
end
real_folders = all_folders(real_folders); %filtered folder names so all voltage levels
folder_num = 1; %set which folder to get values from

pixel_to_um_ratio = get_calibration(directory,calibration_directory,real_folders,folder_num); %take the first folder and access what binning setting used in firing parameters csv sheet
[voltage_vector,current_voltage,resistance,light_pulse,light_step,number_of_firings,resistor_1_pulse] = get_firing_parameters(directory,real_folders,folder_num);

%% Main
if iter == 1
    area_holder_backgroundsubtract = zeros(length(real_folders),number_of_firings,length(real_folders_repeats));
    time_holder = light_step.*(1:number_of_firings);
    voltage_vector_plot = zeros(length(real_folders_repeats),length(voltage_vector)); %accounds for if it is not the right order
end
Hblob = vision.BlobAnalysis();
se = strel('disk',7); %may need to adjust
net = denoisingNetwork('DnCNN');
area_threshold = 50; %pixels

parfor ii = 1:length(real_folders) %go through all folders / voltage levels
    [~,current_voltage,~,~,~,~,~] = get_firing_parameters(directory,real_folders,ii);
    voltage_vector_plot(iter,ii) = current_voltage;

    current_folder = real_folders(ii).name;
    processed_folder = [current_folder,' -- processed'];
    if isfolder(fullfile(directory,processed_folder))
        rmdir(fullfile(directory,processed_folder),'s') %delete the directory if exists
    end
    mkdir(fullfile(directory,processed_folder))
    %rename_tiffs(directory,file) %rename the folder (not needed anymore)
    
    all_images = dir(fullfile(directory,current_folder,'*.tiff'));
    all_images = {all_images.name};
    
    %mean_frame = compute_average_frame(directory,current_folder,net,noise_filter_type); %compute average frame from all images
    mean_frame = imread(fullfile(directory,current_folder,all_images{1})); %use first frame
    if strcmp(noise_filter_type,'deep')
        mean_frame = denoiseImage(im2double(im2gray(mean_frame)),net); %denoise, grayscale, double
    end
    if strcmp(noise_filter_type,'wiener')
        mean_frame = wiener2(im2double(im2gray(mean_frame)),[5,5]); %denoise, grayscale, double
    end

    for jj = 1:number_of_firings %go through all images in folder
        clc
        fprintf('Bubble Extraction and Analysis for V = %.1f: \n',current_voltage)
        disp([num2str(jj*100/number_of_firings),'%'])
        current_image_name = all_images{jj};
        im = imread(fullfile(directory,current_folder,current_image_name));
        im = im2gray(im);
        im2 = im2double(im);
        if strcmp(noise_filter_type,'wiener')
            denoised_im2 = wiener2(im2,[5,5]); %for faster denoising
        end
        if strcmp(noise_filter_type,'deep')
            denoised_im2 = denoiseImage(im2,net); %more robust, but much slower 
        end
        im_sub = imsubtract(mean_frame,denoised_im2); 
        im_sub = imadjust(im_sub,[.2,1],[0,1]); %may need to modify to get correct [.04,1],[0,1]
        im_sub = bwareaopen(im_sub,area_threshold);
        bubble_image_dilate2 = imdilate(im_sub,se);
        bubble_image_open2 = imclose(bubble_image_dilate2,se);
        bubble_image_erode2 = imerode(bubble_image_open2,se);
        bubble_image_filled2 = imfill(bubble_image_erode2,'holes');
        
        [area2,centroid2,bbox2] = Hblob(bubble_image_filled2);
        area_sorted2 = sort(area2,'descend');
        CC2 = bwconncomp(bubble_image_filled2);
        L2 = labelmatrix(CC2);
        B2 = labeloverlay(denoised_im2,L2);
        if size(B2,3) == 1 %no colors so monochrome
            temp = uint8(zeros([size(B2),3]));
            temp(:,:,1) = B2;
            temp(:,:,2) = B2;
            temp(:,:,3) = B2;
            B2 = temp;
        end       
        imwrite(B2,fullfile(directory,processed_folder,current_image_name))      
        
%         figure(1)
%         subplot(2,4,1)
%         imshow(im2)
%         title('Original')
%         subplot(2,4,2)
%         imshow(denoised_im2)
%         title('Denoised')
%         subplot(2,4,3)
%         imshow(im_sub)
%         title('Subtract, Adjusted, Opened')
%         subplot(2,4,4)
%         imshow(bubble_image_dilate2)
%         title('Dilate')
%         subplot(2,4,5)
%         imshow(bubble_image_open2)
%         title('Close')
%         subplot(2,4,6)
%         imshow(bubble_image_erode2)
%         title('Erode')
%         subplot(2,4,7)
%         imshow(bubble_image_filled2)
%         title('Fill')
%         subplot(2,4,8)
%         imshow(B2)
%         title('Label Overlay')
        
        if isempty(area_sorted2) || max(area_sorted2) > numel(im) / 5 || jj <= 2 %last two parts not needed with uniform light
            continue
        else
            % ii --> voltage, jj --> time, iter --> repeat number
            area_holder_backgroundsubtract(ii,jj,iter) = sum(area_sorted2); %take largest area region only (max) vs sum
        end
        
    end
end

end

voltage_vector = voltage_vector_plot(1,:); %voltages to match the order in parallel processing
[ordered_voltage_vector,index] = sort(voltage_vector,'ascend');

figure('units','normalized','outerposition',[0 0 1 1])
hold on
main_max_data = squeeze(max(area_holder_backgroundsubtract,[],2));
for rr = 1:size(main_max_data,1)
    [~,outlier_loc] = rmoutliers(main_max_data(rr,:));
    main_max_data(rr,outlier_loc) = nan();
end
errorbar(voltage_vector',mean(main_max_data,2,'omitnan')./(pixel_to_um_ratio^2),std(main_max_data,0,2,'omitnan')./(pixel_to_um_ratio^2),'ok','MarkerSize',10,'MarkerFaceColor','k','LineWidth',2)
hold off
grid on
set(gca,'GridAlpha',.5)
xlabel('Voltage (V)')
ylabel('Max Bubble Area (\mum^2)')
set(gca,'FontSize',28,'FontWeight','Bold')
saveas(gcf,fullfile(main_directory,'max bubble area vs voltage.png'))

skip_number = 2;
figure('units','normalized','outerposition',[0 0 1 1])
hold on
legend_label = cell(1,length(1:skip_number:length(voltage_vector)));
color_space = distinguishable_colors(length(voltage_vector));
counter = 1;
total_max = max(mean(main_max_data,2,'omitnan'));
for ii = 1:skip_number:length(voltage_vector)
    legend_label{counter} = sprintf('V = %.1f',voltage_vector(index(ii)));
    main_data = squeeze(area_holder_backgroundsubtract(index(ii),:,:));
    for rr = 1:size(main_data,1) %for each time step
        outlier_loc = (main_data(rr,:) + 3.*std(main_data(rr,:))) > 1.5*total_max;
        main_data(rr,outlier_loc) = nan();
        %[~,outlier_loc] = rmoutliers(main_data(rr,:));
        %main_data(rr,outlier_loc) = nan();
    end
    mean_data = mean(main_data,2,'omitnan');
    std_data = std(main_data,0,2,'omitnan');
    errorbar(time_holder,mean_data./(pixel_to_um_ratio^2),std_data./(pixel_to_um_ratio^2),'.','MarkerSize',10,'Color',color_space(index(ii),:))
    counter = counter + 1;
end
hold off
xlabel('Time (\mus)')
ylabel('Bubble Area (\mum^2)')
set(gca,'FontSize',28,'FontWeight','Bold')
legend(legend_label,'Location','EastOutside','FontSize',22)
grid on
set(gca,'GridAlpha',.5)
%xlim([0,70]) %set depending on the specific case
%ylim([0,2.5e5])
saveas(gcf,fullfile(main_directory,'bubble area vs time.png'))

toc

%% Auxillary Functions

function mean_frame = compute_average_frame(directory,current_folder,net,noise_filter_type)

all_images = dir(fullfile(directory,current_folder,'*.png'));
all_images = {all_images.name};
for jj = 1:length(all_images)
    clc
    fprintf('Computing Mean Frame: \n')
    disp([num2str(jj*100/length(all_images)),'%'])
    current_image_name = all_images{jj};
    im = imread(fullfile(directory,current_folder,current_image_name));
    im = im2gray(im);
    im = im2double(im);
    if jj == 1
        mean_frame = zeros(size(im));
    end
    if strcmp(noise_filter_type,'deep')
        denoised_im = denoiseImage(im,net); 
    end
    if strcmp(noise_filter_type,'wiener')
        denoised_im = wiener2(im,[5,5]);
    end
    mean_frame = mean_frame + denoised_im;
end
mean_frame = mean_frame ./ length(all_images);

end

function rename_tiffs(directory,file)
    
all_files = dir(fullfile(directory,file,'*.tiff'));
all_names = {all_files.name};

for ii = 1:length(all_names)
    current_name = all_names{ii};
    total_length = length(num2str(length(all_names)));
    underline_locs = strfind(current_name,'_');
    tiff_loc = strfind(current_name,'.tiff');
    num = current_name(underline_locs(end)+1:tiff_loc - 1);
    num_length = length(num2str(num));
    zero_repeat = repmat('0',1,total_length - num_length);
    new_number = [zero_repeat,num];
    new_name = ['image_',new_number,'.tiff'];
    try
    movefile(fullfile(directory,file,current_name),fullfile(directory,file,new_name)); %rename
    end
end
end

function pixel_to_um_ratio = get_calibration(directory,calibration_directory,real_folders,folder_num)

csv_file = dir(fullfile(directory,real_folders(folder_num).name,'*.csv'));
raw = readtable(fullfile(directory,real_folders(folder_num).name,csv_file.name),'PreserveVariableNames', true);
csv_labels = raw.Properties.VariableNames; %get all string labels in cell array format
bin_index = 1;
for ii = 1:length(csv_labels)
    if strfind(csv_labels{ii},'binning') %find the binning index
        bin_index = ii;
        break
    end
end
binning_value = raw.(csv_labels{bin_index}); %get binning value

if binning_value == 1 %load in the calibration image based on which binning used
    calibration_image = imread(fullfile(calibration_directory,'calibration 0.15 mm diameter circle, binning 1.tiff'));
elseif binning_value == 2
    calibration_image = imread(fullfile(calibration_directory,'calibration 0.15 mm diameter circle, binning 2.tiff'));
else %default to binning = 2
    calibration_image = imread(fullfile(calibration_directory,'calibration 0.15 mm diameter circle, binning 2.tiff'));
end 



bw = imbinarize(calibration_image,.75); %convert to bw for regionprops
stats = regionprops(bw,'EquivDiameter');
diameter_in_pixels = stats.EquivDiameter;
actual_diameter = 150; %um

pixel_to_um_ratio = diameter_in_pixels./actual_diameter;

end

function [voltage_vector,current_voltage,resistance,light_pulse,light_step,number_of_firings,resistor_1_pulse] = get_firing_parameters(directory,real_folders,folder_num)

csv_file = dir(fullfile(directory,real_folders(folder_num).name,'*.csv'));
raw = readtable(fullfile(directory,real_folders(folder_num).name,csv_file.name),'PreserveVariableNames', true);
csv_labels = raw.Properties.VariableNames; %get all string labels in cell array format

voltage_start_index = 1;
current_voltage_index = 1;
resistance_index = 1;
light_pulse_index = 1;
light_step_index = 1;
number_of_firings_index = 1;
resistor_1_pulse_index = 1;

for ii = 1:length(csv_labels)
    if strfind(csv_labels{ii},'voltage start') %find voltage start index
        voltage_start_index = ii;
    end
    if strfind(csv_labels{ii},'voltage (V)') %find the voltage index
        current_voltage_index = ii;
    end    
    if strfind(csv_labels{ii},'resistance') %find the resistance index
        resistance_index = ii;
    end     
    if strfind(csv_labels{ii},'light pulse') %find the light pulse index
        light_pulse_index = ii;
    end        
    if strfind(csv_labels{ii},'light step') %find the light step index
        light_step_index = ii;
    end     
    if strfind(csv_labels{ii},'number of firings') %find the light step index
        number_of_firings_index = ii;
    end      
    if strfind(csv_labels{ii},'resistor 1 pulse') %find the light step index
        resistor_1_pulse_index = ii;
    end        
end

voltage_start = raw.(csv_labels{voltage_start_index}); %get value
voltage_step = raw.(csv_labels{voltage_start_index+1}); %get value
voltage_stop = raw.(csv_labels{voltage_start_index+2}); %get value
voltage_vector = voltage_start:voltage_step:voltage_stop;

current_voltage = raw.(csv_labels{current_voltage_index}); %get value
resistance = raw.(csv_labels{resistance_index}); %get value
light_pulse = raw.(csv_labels{light_pulse_index}); %get value
light_step = raw.(csv_labels{light_step_index}); %get value
number_of_firings = raw.(csv_labels{number_of_firings_index}); %get value
resistor_1_pulse = raw.(csv_labels{resistor_1_pulse_index}); %get value

end
